# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from .magic_extension import *
from .version import VERSION
from .kql_magic import kql 


__version__ = VERSION

